{********************************************************************}
{                                                                    }
{ written by TMS Software                                            }
{            copyright (c) 2019                                      }
{            Email : info@tmssoftware.com                            }
{            Web : http://www.tmssoftware.com                        }
{                                                                    }
{ The source code is given as is. The author is not responsible      }
{ for any possible damage done due to the use of this code.          }
{ The complete source code remains property of the author and may    }
{ not be distributed, published, given or sold in any form as such.  }
{ No parts of the source code can be included in any other component }
{ or application without written authorization of the author.        }
{********************************************************************}
unit WEBLib.IndexedDb;

interface

uses
  {$IFNDEF LCLTMSWEB}
  System.Classes, DB, DBClient, WEBLib.Controls, Web, jsdelphisystem, js;
  {$ENDIF}
  {$IFDEF LCLTMSWEB}
  Classes, DB, WEBLib.Controls, Web, js, jsdelphisystem;
  {$ENDIF}

type
  TIndexedDbOpCode = (opOpen, opAdd, opPut, opDelete, opGet, opGetAllKeys, opGetAllObjs,
                      opGetAllIndexKeys, opGetAllObjsByIndex);

  TIndexedDbResultEvent = procedure(success: Boolean; opCode: TIndexedDbOpCode; data: JSValue; sequenceId: JSValue; errorName, errorMsg: String) of object;

  TIndexedDb = class(TComponent)
  private
     FDatabaseName: String;
     FobjectStoreName: string;
     FKeyFieldName: String;
     FAutoIncrement: Boolean;
     FActiveIndex: String;
     FIndexDescending: Boolean;

     FOnResult: TIndexedDbResultEvent;
  protected
  public
    constructor Create(AOwner: TComponent); overload; override;
    destructor Destroy; override;

    // Add Indexes before opening a "new" database. If database exists then
    // AddIndex will be ignored.
    procedure AddIndex(anIndexName: String; fields: String; isUnique: Boolean=False);

    { Explanation of parameters:
     Pass a DBName and ObjectStoreName. These will be created if absent.
     KeyFieldName: specify if you want to use inline-key (data contains key). In that case, don't pass key separately to PutData below.
                   The key field value will then be taken from that named property from the object passed to PutData.
     AutoIncrement: This will be effective only if db does not exist so that the db is created with autoincrement.
                    In case of autoincrement, you need not pass key to addData. The auto-created key is returned
                    in its success result.
    }
    procedure Open(aDbName: String; aObjectStoreName: String; KeyFieldName: String = ''; autoIncrement: Boolean = False; sequenceID: Integer = 0);

    procedure PutData(akey: JSValue; data: JSValue; sequenceID: Integer = 0); overload;
    procedure PutData(data: JSValue; sequenceID: Integer = 0); overload;
    procedure AddData(data: JSValue; sequenceID: Integer = 0);
    procedure DeleteData(AKey: JSValue; sequenceID: Integer = 0);

    procedure GetData(akey: JSValue; sequenceID: Integer = 0);
    procedure GetKeys(sequenceID: Integer = 0);
    procedure GetAllObjs(sequenceID: Integer = 0);
    procedure GetIndexKeys(indexPropertyName: String; sequenceID: Integer = 0);
    // Gets objects by sequence set in properties ActiveIndex, IndexDescending
    // If ActiveIndex is empty then gets by primary index.
    procedure GetAllObjsByIndex(sequenceID: Integer = 0);
    procedure GetIndexData(indexPropertyName: String; akey: JSValue; sequenceID: Integer = 0);

    property DatabaseName: String read FDatabaseName;
    property ObjectStoreName: String read FObjectStoreName;
    property KeyFieldName: String read FKeyFieldName;
    property AutoIncrement: Boolean read FAutoIncrement;
    property ActiveIndex: String read FActiveIndex write FActiveIndex;
    property IndexDescending: Boolean read FIndexDescending write FIndexDescending;
    property OnResult: TIndexedDbResultEvent read FOnResult write FOnResult;
  end;

  TIDBErrorEvent = procedure(DataSet: TDataSet; opCode: TIndexedDbOpCode; errorName, errorMsg: String) of object;
  TIDBInitSuccessEvent = procedure(Sender: TObject) of Object;

  {$IFNDEF LCLTMSWEB}
  TIDBProc = reference to procedure;
  {$ENDIF}
  {$IFDEF LCLTMSWEB}
  TIDBProc = procedure;
  {$ENDIF}

  {$IFNDEF LCLTMSWEB}
  [ComponentPlatforms(TMSWebPlatform)]
  {$ENDIF}
  TWebIndexedDbClientDataset = class(TDataSet)
  private
    FIDBDatabaseName, FIDBObjectStoreName, FIDBKeyFieldName: String;
    FIDBAutoIncrement, FIDBDataLoaded: Boolean;
    FIDBIndexName: String;
    FIDBIndexDescending: Boolean;
    FOnIDBError: TIDBErrorEvent;
    FOnInitSuccess: TIDBInitSuccessEvent;
    FParams: TParams;
    procedure SetParams(const Value: TParams);
  protected
    procedure InternalClose; override;
    procedure InternalHandleException; override;
    procedure InternalInitFieldDefs; override;
    procedure InternalOpen; override;
    function IsCursorOpen: Boolean; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Refresh;

    // Calling Init is necessary only if you have nultiple CDS
    // components using the same database.
    // In such a case, before making those CDS components Active
    // * call Init on the first CDS component.
    // * Pass an event handler to wait for success event before calling
    //   Init on the next CDS component and so on.
    // * Also, if you are using AddIDBIndex,
    //   please make sure that it is called before Init.
    function Init(ProcInitSuccess: TIDBProc=nil): Boolean;

    // Call AddIDBIndex before Init and Set Active calls.
    // Index will be added only for a new object store. If
    // the object store already exists then AddIDBIndex will be
    // ignored.
    procedure AddIDBIndex(anIndexName: String; AFields: String; isUnique: Boolean=False);

    property IDBDataLoaded: Boolean read FIDBDataLoaded write FIDBDataLoaded;

    property IDBActiveIndex: String read FIDBIndexName write FIDBIndexName;
    property IDBIndexDescending: Boolean read FIDBIndexDescending write FIDBIndexDescending;
    // The following can not be published as it is a reference proc
  published
    property Active;
    // Database and ObjectStore will be created if absent
    property IDBDatabaseName: String read FIDBDatabaseName write FIDBDatabaseName;
    property IDBObjectStoreName: String read FIDBObjectStoreName write FIDBObjectStoreName;
    // Specify primary key here
    property IDBKeyFieldName: String read FIDBKeyFieldName write FIDBKeyFieldName;
    // Specify AutoIncrement for above key.
    // Only effective when creating a new db. This can not be used
    // to change AutoIncrement property of an existing DB.
    property IDBAutoIncrement: Boolean read FIDBAutoIncrement write FIDBAutoIncrement;
    property Params: TParams read FParams write SetParams;
    property OnIDBError: TIDBErrorEvent read FOnIDBError write FOnIDBError;
    property OnInitSuccess: TIDBInitSuccessEvent read FOnInitSuccess write FOnInitSuccess;

    // Other
    property BeforeOpen;
    property AfterOpen;
    property BeforeClose;
    property AfterClose;
    property BeforeInsert;
    property AfterInsert;
    property BeforeEdit;
    property AfterEdit;
    property BeforePost;
    property AfterPost;
    property BeforeCancel;
    property AfterCancel;
    property BeforeDelete;
    property AfterDelete;
    property BeforeScroll;
    property AfterScroll;
    property OnCalcFields;
    property OnDeleteError;
    property OnEditError;
    property OnFilterRecord;
    property OnNewRecord;
    property OnPostError;    
  end;

implementation

constructor TIndexedDb.Create(AOwner: TComponent);
begin
end;

destructor TIndexedDb.Destroy;
begin
end;

procedure TIndexedDb.AddIndex(anIndexName: String; fields: String; isUnique: Boolean=False);
begin
end;

procedure TIndexedDb.Open(aDbName: String; aObjectStoreName: String; KeyFieldName: String = ''; autoIncrement: Boolean = False; sequenceID: Integer = 0);
begin
end;

procedure TIndexedDb.PutData(akey: JSValue; data: JSValue; sequenceID: Integer = 0);
begin
end;

procedure TIndexedDb.PutData(data: JSValue; sequenceID: Integer = 0);
begin
end;

procedure TIndexedDb.AddData(data: JSValue; sequenceID: Integer = 0);
begin
end;

procedure TIndexedDb.DeleteData(AKey: JSValue; sequenceID: Integer = 0);
begin
end;

procedure TIndexedDb.GetData(akey: JSValue; sequenceID: Integer = 0);
begin
end;

procedure TIndexedDb.GetKeys(sequenceID: Integer = 0);
begin
end;

procedure TIndexedDb.GetAllObjs(sequenceID: Integer = 0);
begin
end;

procedure TIndexedDb.GetIndexKeys(indexPropertyName: String; sequenceID: Integer = 0);
begin
end;

procedure TIndexedDb.GetAllObjsByIndex(sequenceID: Integer = 0);
begin
end;

procedure TIndexedDb.GetIndexData(indexPropertyName: String; akey: JSValue; sequenceID: Integer = 0);
begin
end;

procedure TWebIndexedDbClientDataset.InternalClose;
begin
  //
end;

procedure TWebIndexedDbClientDataset.InternalHandleException;
begin
  //
end;

procedure TWebIndexedDbClientDataset.InternalInitFieldDefs;
begin
  //
end;

procedure TWebIndexedDbClientDataset.InternalOpen;
begin
  //
end;

function TWebIndexedDbClientDataset.IsCursorOpen: Boolean;
begin
  Result := false;
end;

constructor TWebIndexedDbClientDataset.Create(AOwner: TComponent);
begin
  inherited;
  FIDBDatabaseName := '';
  FIDBObjectStoreName := '';
  FIDBKeyFieldName := '';
  FIDBAutoIncrement := True;
  FParams := TParams.Create(Self);
end;

destructor TWebIndexedDbClientDataset.Destroy;
begin
  FParams.Free;
  inherited;
end;

procedure TWebIndexedDbClientDataset.Refresh;
begin
end;

procedure TWebIndexedDbClientDataset.SetParams(const Value: TParams);
begin
  FParams.Assign(Value);
end;

procedure TWebIndexedDbClientDataset.AddIDBIndex(anIndexName: String; AFields: String; isUnique: Boolean=False);
begin
end;

function TWebIndexedDbClientDataset.Init(ProcInitSuccess: TIDBProc=nil): Boolean;
begin
  Result := False;
end;

end.
